// C++
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

// OpenGL
#include <GL/glut.h>

// Local
#include <cyl.h>
#include <Matrix4.h>
#include <Vector3.h>
#include <Vector4.h>

using namespace std;

const int SUBDIV_MODE_STAR = 0;
const int SUBDIV_MODE_TRI = 0;

Cyl::Cyl ()
{
  Shape ();
  solid = Shape::SOL_CYL;
  radius = 2.0;
  //height = 2.0;
  // circ = 2 * height, so can fit texture image of 1:2 ratio w/o stretching
  height = BasicMath::getPi() * radius; // * 2 / 2
  doTexture = false;
  cx = 0;
  cy = 0;
  cz = 0;
}

Cyl::Cyl (int s)
{
  Shape ();
  solid = s;
  radius = 1.0;
  height = 2.0;
  doTexture = false;
  cx = 0;
  cy = 0;
  cz = 0;
}

void Cyl::setCenter (float x, float y, float z)
{
  cx = x;
  cy = y;
  cz = z;
}

void Cyl::setRadius (float r)
{
  radius = r;
}

void Cyl::setHeight (float h)
{
  height = h;
}

void Cyl::setTex (bool t)
{
  doTexture = t;
}

// Draw an icosahedron, a sphere approximation with 20 triangles and 12 verts,
//   where each face is an equalateral triangle.
// TODO TODO: change these arrays of floats into our own Vertex3 type!!!!!
void Cyl::draw ()
{
  // Red book uses some magic numbers so that dist from Origin to any vert is
  //   1, i.e. radius 1. We don't need that.
  const float X = 5;
  const float Y = 5;
  const float Z = 5;

  const int NUM_VERTS = 8;
  // Define positions of each vertex (points)
  static float vertData_arr [NUM_VERTS][3] = {
    {-X, Y, Z}, {X, Y, Z}, {X, -Y, Z}, {-X, -Y, Z},
    {-X, Y, -Z}, {X, Y, -Z}, {X, -Y, -Z}, {-X, -Y, -Z}
  };

  Vector3 vertData [NUM_VERTS];
  // Move array data into Vector3 type data
  for (int i = 0; i < NUM_VERTS; i++)
  {
    // Move each arr pt to a Vector3 pt
    vertData [i] = Vector3 (vertData_arr [i][0], vertData_arr[i][1],
      vertData_arr [i][2]);
  }

  const int NUM_TRIS = 6;
  // Define the indices of vertData pts that can make a triangle (triangles)
  // These indices refer to the index in vertData array
  // Must be CCW so that faces show up when enable backface culling.
  static int triIndices [NUM_TRIS][4] = {
    {0,1,2,3}, {5,4,7,6}, {1,5,6,2}, {4,0,3,7}, {4,5,1,0}, {3,2,6,7}
  };

  // For calc-ing normals of the actual icosahedron, not the approx sphere
  //Vector3 diff1 = Vector3 ();
  //Vector3 diff2 = Vector3 ();
  //Vector3 crossed = Vector3 ();

  // Reset
  num_faces = 0;

  // Draw quads in a loop, each iteration draws 1 triangle
  for (int i = 0; i < NUM_TRIS; i ++)
  {
    // Subdivide each triangle into 4 smaller triangles, and draw them
    subdivide (vertData [triIndices [i] [0]],
      vertData [triIndices [i] [1]],
      vertData [triIndices [i] [2]],
      //2, SUBDIV_MODE_TRI);

      // to test 6-face cube
      vertData [triIndices [i] [3]],
      0, SUBDIV_MODE_TRI);
  }

  //fprintf (stderr, "Drew %d faces total.\n", num_faces);

}


// Subdivide 1 quad into multiple ones so have finer granularity to
//   approximate a shape
//   Calculate the midpoints of each of the 4 edges. These 4 midpts, together
//   with the center of the quad, form a new quad, which divides the orig
//   quad into 4 equalateral smaller ones.
// Parameters:
//   v1: 1st vertex's coordinates. Array of 3 floats
//   v2: 2nd vertex's coordinates. Array of 3 floats
//   v3: 3rd vertex's coordinates. Array of 3 floats
//   v4: 4th vertex's coordinates. Array of 3 floats
// Big quad divided into 4 smaller ones, by clipping at the midpt of each
//   of the 4 edges and the center pt of the quad:
//   _______
//   |  |  |
//   |__|__|
//   |  |  |
//   |__|__|
//
void Cyl::subdivide (Vector3& v1, Vector3& v2, Vector3& v3, Vector3& v4,
  int depth, int mode)
{
  // Init midpts. midpts btwn edges 1 & 2, btwn edges 2 & 3, btwn edges 3 & 1
  //   (Numbers in names are given in CCW direction, to be consistent.)
  Vector3 v12 = Vector3 ();
  Vector3 v23 = Vector3 ();
  Vector3 v34 = Vector3 ();
  Vector3 v41 = Vector3 ();
  Vector3 v5 = Vector3 ();

  // Normalize the orig ones, in case they're not magnitude 1
  v1.norm (v1);
  v2.norm (v2);
  v3.norm (v3);
  v4.norm (v4);

  // Lower the depth, means smaller the triangle. When depth is 0, that means
  //   we've divided enough number of times that we are at the base triangle.
  //   Can draw it.
  if (depth == 0)
  {
    glBegin (GL_TRIANGLES);
      // cylinder
      if (solid == SOL_CYL)
      {
        // This translates, then when rotate, will rotate around origin and
        //   Orbit. If don't want that, draw it at origin!
        /*
        drawCirc (1.0, 0.0, 10.0, 1.0, 0.0, 5, Shape::PATH_MODE_CIRC,
          SOL_CYL, 2);
        */

        drawCirc (cx, cy, cz, radius, 0.0, 10, Shape::PATH_MODE_CIRC,
          SOL_CYL, height, doTexture);
        //drawCirc (0.0, 0.0, 0.0, 2.0, 0.0, 10, Shape::PATH_MODE_CIRC, 0);
        //drawCirc (0.0, 0.0, 2.0, 2.0, 0.0, 10, Shape::PATH_MODE_CIRC, 0);
      }

      // cone
      else if (solid == SOL_CONE)
        drawCirc (6.0, 0.0, 5.0, radius, 0.0, 5, Shape::PATH_MODE_CIRC,
          SOL_CONE, height, doTexture);
    glEnd ();
    return;
  }

  // loop through the (x y z) coords of each pt
  if (mode == SUBDIV_MODE_STAR)
  {
    v12 = v1 + v2;
    v12.mul (0.5, v12);
    v23 = v2 + v3;
    v23.mul (0.5, v23);
    v34 = v3 + v4;
    v34.mul (0.5, v34);
    v41 = v4 + v1;
    v41.mul (0.5, v41);

    // Diagonal pts added, then / 2, gives the center of cube
    v5 = v1 + v3;
    v5.mul (0.5, v5);
  }

  // Normalize the 3 midpts if want radius 1. Else don't need to?
  v12.norm (v12);
  v23.norm (v23);
  v34.norm (v34);
  v41.norm (v41);
  v5.norm (v5);

  // Recursive call for each of the 4 subdivided quads
  subdivide (v1, v12, v5, v41, depth-1, mode);
  subdivide (v12, v2, v23, v5, depth-1, mode);
  subdivide (v5, v23, v3, v34, depth-1, mode);
  subdivide (v41, v5, v34, v4, depth-1, mode);

}

