// C++
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

// OpenGL
#include <GL/glut.h>

// Local
#include <shape.h>
#include <Matrix4.h>
#include <Vector3.h>
#include <Vector4.h>

using namespace std;

const int SUBDIV_MODE_STAR = 0;
const int SUBDIV_MODE_TRI = 1;

Shape::Shape()
{
  angle = 0.0;
  num_faces = 0;
  _depth = 2; // default depth 2, 320 triangles

  _scalex = 1;
  _scaley = 1;
  _scalez = 1;

  _rotx = 1;
  _roty = 1;
  _rotz = 1;
  _rotw = 1;

  _tex_w = 0;
  _tex_h = 0;

  matrix = Matrix4 ();
  _mat_rot = Matrix4 ();
  _mat_rot.identity ();
  _mat_scale = Matrix4 ();
  _mat_scale.identity ();
  _mat_trans = Matrix4 ();
  _mat_trans.identity ();
}

Matrix4& Shape::getMatrix()
{
  return matrix;
}

void Shape::setMatrix (Matrix4 mat)
{
  for (int i = 0; i < 16; i++)
    //for (int j = 0; j < 4; j++)
      //matrix.set (i,j,mat.get(i,j));
    matrix.set (i,mat.get(i));
}

void Shape::setScale (double sx, double sy, double sz)
{
  _scalex = sx;
  _scaley = sy;
  _scalez = sz;
}

void Shape::setRotate (double rx, double ry, double rz, double rw)
{
  _rotx=rx;
  _roty=ry;
  _rotz=rz;
  _rotw=rw;
}

// Set params to use when drawing vertices in drawVert(), applied directly on
//   to the verts
void Shape::setMulRotate (double rx, double ry, double rz, double rw)
{
  Vector4 newrot = Vector4 (rx, ry, rz, rw);
  Vector4 res = Vector4 ();

  Matrix4 mat = Matrix4 ();
  mat.rotate (_rotx, _roty, _rotz, _rotw);

  // Mult by new params to rotate, store into newvec
  mat.mul (newrot, res);

  // Store results into class vars so can rotate each vertex when drawVert
  _rotx = res[0];
  _roty = res[1];
  _rotz = res[2];
  _rotw = res[3];
}

void Shape::setDepth (int d)
{
  _depth = d;
}

void Shape::updateMat ()
{
  matrix.identity ();
  /*
  _mat_scale.mul (matrix, matrix);
  _mat_rot.mul (matrix, matrix);
  _mat_trans.mul (matrix, matrix);
  */
  
  matrix.mul (_mat_scale, matrix);
  matrix.mul (_mat_rot, matrix);
  matrix.mul (_mat_trans, matrix);
  
}

void Shape::setTexDim (int w, int h)
{
  _tex_w = w;
  _tex_h = h;
}

// Angle is accumulative
void Shape::spin(double deg)
{
  angle += deg;
  if (angle > 360.0 || angle < -360.0) angle = 0.0;

  _mat_rot.rotateY (angle);

  // Update matrix
  updateMat ();
}

// Around arbitrary axis
void Shape::spin (double ax, double ay, double az, double deg)
{
  angle += deg;
  if (angle > 360.0 || angle < -360.0) angle = 0.0;

  _mat_rot.rotate (ax,ay,az,angle);
  updateMat ();
}

// Parameters:
//   r*: x y z coordinates of the axis wrt which to rotate
void Shape::rotate (double rx, double ry, double rz, double ang)
{
  // Rotate shape's rotational mat
  _mat_rot.rotate (rx, ry, rz, ang);

  // update full mat
  updateMat ();
}

// Set the local matrix (not the actual obj vertices, the local coord sys)
// Parameters:
//   r*: x y z coordinates of the axis wrt which to rotate
void Shape::mulRotate (double rx, double ry, double rz, double ang)
{
  // Make a new rotational mat
  Matrix4 mat1 = Matrix4 ();
  mat1.rotate (rx, ry, rz, ang);
  //Debug //mat1.rotate (1, 1, 1, 1);

  // Test
  // Pass in ang == 0, so this would be I * I = I, which shouldn't change anyth
  // iden
  //Debug //_mat_rot = Matrix4 (1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1);
  //Debug //_mat_rot.print ();

  // multiply to shape's existing mat
  _mat_rot.mul (mat1, _mat_rot);
  // update full mat
  updateMat ();
}

void Shape::scale (double sx, double sy, double sz)
{
  _mat_scale.scale (sx, sy, sz);

  // Update matrix
  updateMat ();
}

void Shape::mulScale (double sx, double sy, double sz)
{
  Matrix4 mat1 = Matrix4 ();
  mat1.scale (sx, sy, sz);
  _mat_scale.mul (mat1, _mat_scale);
  updateMat ();
}

void Shape::trans (double tx, double ty, double tz)
{
  _mat_trans.translate (tx, ty, tz);
  updateMat ();
}

void Shape::mulTrans (double tx, double ty, double tz)
{
  Matrix4 mat1 = Matrix4 ();
  mat1.translate (tx, ty, tz);
  _mat_trans.mul (mat1, _mat_trans);
  updateMat ();
}



// ------------------------------------------------------------ drawing fns --

// Parameters:
//   colors: array specifying color for each vertex.
//   solidColor: if colors param is NULL (i.e. don't have a array specifying
//     color for each vertex, can specify 1 solid color for the obj.
//     If this is also NULL, then will assign colors by default (i.e. by the
//     number of polygon.)
void Shape::doDraw (int nVerts, float * verts, int nIndices,
  int * indices, float * colors, float * solidColor, float * normals,
  float * texcoords)
{
  Vector3 p1 = Vector3 (), p2 = Vector3 (), p3 = Vector3 ();
  Vector3 color1 = Vector3 (), color2 = Vector3 (), color3 = Vector3 (),
    color = Vector3 ();
  // texture
  Vector3 t1 = Vector3 (), t2 = Vector3 (), t3 = Vector3 ();

  int vidx = 0;
  int coordIdx = 0;

  glBegin (GL_TRIANGLES);
  // Draw a triangle in each iter
  //   Set up each of the 3 pts in the triangle, then draw it
  for (int i = 0; i < nIndices; i++)
  {
    vidx = indices [i];
    coordIdx = 3 * vidx;
    p1 = Vector3 (verts [coordIdx],
      verts [coordIdx + 1],
      verts [coordIdx + 2]);
    if (texcoords != NULL)
    {
      t1.setAll (texcoords [coordIdx],
        texcoords [coordIdx + 1],
        texcoords [coordIdx + 2]);
    }
    else if (colors != NULL)
    {
      color1 = Vector3 (colors [coordIdx],
        colors [coordIdx + 1],
        colors [coordIdx + 2]);
    }

    i++;
    if (i >= nIndices)
      break;
    vidx = indices [i];
    coordIdx = 3 * vidx;
    p2 = Vector3 (verts [coordIdx],
      verts [coordIdx + 1],
      verts [coordIdx + 2]);
    if (texcoords != NULL)
    {
      t2.setAll (texcoords [coordIdx],
        texcoords [coordIdx + 1],
        texcoords [coordIdx + 2]);
    }
    else if (colors != NULL)
    {
      color2 = Vector3 (colors [coordIdx],
        colors [coordIdx + 1],
        colors [coordIdx + 2]);
    }

    i++;
    if (i >= nIndices)
      break;
    vidx = indices [i];
    coordIdx = 3 * vidx;
    p3 = Vector3 (verts [coordIdx],
      verts [coordIdx + 1],
      verts [coordIdx + 2]);
    if (texcoords != NULL)
    {
      t3.setAll (texcoords [coordIdx],
        texcoords [coordIdx + 1],
        texcoords [coordIdx + 2]);
    }
    else if (colors != NULL)
    {
      color3 = Vector3 (colors [coordIdx],
        colors [coordIdx + 1],
        colors [coordIdx + 2]);
    }

    if (texcoords != NULL)
    {
      drawTri_tex (p1,p2,p3,  t1,t2,t3);
    }
    else if (colors != NULL)
      drawTri_clr (p1, p2, p3, color1, color2, color3);
    else if (solidColor != NULL)
    {
      color = Vector3 (solidColor [0], solidColor [1], solidColor [2]);
      drawTri_clr (p1, p2, p3, color, color, color);
    }
    else
      drawTri (p1, p2, p3);

    //p1.print ();
    //p2.print ();
    //p3.print ();
  }
  glEnd ();

}



/*
// cx: x-coord of center of circ to draw
// cy: z-coord of center of circ to draw
// r: radius of circle, or a of ellipse, if mode is specified for ellp
// b: b axis length of ellipse, if mode is specified for ellp
// x_step: How fine grain the calculation needs to be. 0 to 360
// mode: constant, circ or ellp
void Shape::drawCirc_old (double cx, double cz, double r, double b,
  int x_step, int mode)
{
  // To draw triangles that form each sector of the circle
  //   v1 is always the center of the circle. v2 and v3 go CCW, so the RHS
  //   and LHS verts on the isoceles triangle's base, respectively
  Vector3 v1 = Vector3 (cx, cz, 0.0);
  Vector3 v2, v3;
  Vector3 v_init;

  float x = 0, z = 0;
  
  const int TTL_FRMS = 360; 
  const int QUARTER_FRM = TTL_FRMS / 4;
      
  // Divide radius length into QUARTER_FRM slices
  float r_slice = r / QUARTER_FRM;
  int curr_frm;

  int last_term = TTL_FRMS - x_step + 1;
  

  for (int i = 0; i < last_term; i+=x_step)
  {

    // curr_frame should be whatever the current timestep mod TTL_FRMS, so
    //   curr_frame is always btwn 0 and TTL_FRMS-1, inc by 1 each fn call (`.`
    //   timestep inc by 1 each fn call)
    //curr_frm = i % TTL_FRMS; 
    curr_frm = i; 
    
    // Check for 4 quarters of the circle, see which part we are on
    if ((float) curr_frm / TTL_FRMS < 0.25)
    {
      // starting from postivie x axis, x should be max of radius
      x = r_slice * (QUARTER_FRM - (curr_frm % QUARTER_FRM));
  
      if (mode == PATH_MODE_CIRC)
        // eqn of circle: (x-cx)^2 + (z-cz)^2 = r^2
        // Calc z from all other known vars
        z = sqrt (fabs (r*r - ((x-cx) * (x-cx))) ) + cz;
      else if (mode == PATH_MODE_ELLP)
        // eqn of ellipse: ((x-cx)^2 / a^2) + ((z-cz)^2 / b^2) = 1
        z = sqrt (fabs ((1 - ((x-cx)*(x-cx) / (r*r))) ) * b*b) + cz;
    }
    else if ((float) curr_frm / TTL_FRMS < 0.5)
    {
      x = r_slice * (curr_frm % QUARTER_FRM);
      x *= -1;
      if (mode == PATH_MODE_CIRC)
        z = sqrt (fabs (r*r - ((x-cx) * (x-cx))) ) + cz;
      else if (mode == PATH_MODE_ELLP)
        z = sqrt (fabs ((1 - ((x-cx)*(x-cx) / (r*r))) ) * b*b) + cz;
    }
    else if ((float) curr_frm / TTL_FRMS < 0.75)
    {
      x = r_slice * (QUARTER_FRM - (curr_frm % QUARTER_FRM));
      x *= -1;
      if (mode == PATH_MODE_CIRC)
        z = sqrt (fabs (r*r - ((x-cx) * (x-cx))) ) + cz;
      else if (mode == PATH_MODE_ELLP)
        z = sqrt (fabs ((1 - ((x-cx)*(x-cx) / (r*r))) ) * b*b) + cz;
      z *= -1;
    }
    else if ((float) curr_frm / TTL_FRMS < 1)
    {
      x = r_slice * (curr_frm % QUARTER_FRM);
      if (mode == PATH_MODE_CIRC)
        z = sqrt (fabs (r*r - ((x-cx) * (x-cx))) ) + cz;
      else if (mode == PATH_MODE_ELLP)
        z = sqrt (fabs ((1 - ((x-cx)*(x-cx) / (r*r))) ) * b*b) + cz;
      z *= -1;
    }

    // For even i, set v3, `.` LHS vert is calculated before RHS, in the CW
    //   order we're calculating. But verts need to be in CCW order. So set
    //   v3 first, odd i set v2.
    // y-coord == 0 `.` draw circ on the XZ plane
    if (i == 0)
    {
      v3.set (0, x);
      v3.set (1, z);
      v3.set (2, 0);

      // Save for last triangle
      v_init.set (0, x);
      v_init.set (1, z);
      v_init.set (2, 0);
    }

    // Draw normally for all other triangles. Set LHS base vert v3 to be same
    //   as prev triangle's RHS base vert v2
    else
    {
      // 1st iteration will draw the triangle with v3 computed in 0th iteratn,
      //   so don't reset v3. All later triangles' v3 should be the v2 of the
      //   previous triangle.
      // Assumption: v2 values have not changed since the prev triangle in
      //   last iter.
      if (i != x_step)
      {
        v3.set (0, v2 [0]);
        v3.set (1, v2 [1]);
        v3.set (2, v2 [2]);
      }

      if  (i != last_term - x_step)
      {
        v2.set (0, x);
        v2.set (1, z);
        v2.set (2, 0);
      }
      // For last iteration, draw the last triangle closing in, RHS vert
      //   overlapping w/ the LHS vert of the 0th triangle.
      else
      {
        v2.set (0, v_init[0]);
        v2.set (1, v_init[1]);
        v2.set (2, v_init[2]);
      }

      drawTri (v1, v2, v3);

    }
  }
}
*/


// Draws circ, if pass in argument isCyl, can draw cylinder and cone.
// cx: x-coord of center of circ to draw
// cy: z-coord of center of circ to draw
// r: radius of circle, or a of ellipse, if mode is specified for ellp
// b: b axis length of ellipse, if mode is specified for ellp
// x_step: How fine grain the calculation needs to be. 0 to 360
// mode: constant, circ or ellp
// Does NOT implement ellps drawing yet. Use circPath (ELLP) to get ellp path.
// isCyl: 1 = cyl, will draw 2 triangles (=quad) for each height face of cyl;
//        2 = cone, will draw 1 triangle for each height face of cone.
// solHeight: if isCyl is not 0, supply a height to draw the solid
void Shape::drawCirc (double cx, double cy, double cz, double r, double b,
  int x_step, int mode, int isCyl, float solHeight, bool doTexture)
{
  // To draw triangles that form each sector of the circle
  //   v1 is always the center of the circle. v2 and v3 go CCW, so the RHS
  //   and LHS verts on the isoceles triangle's base, respectively
  Vector3 v1 = Vector3 (cx, cy, cz);
  //fprintf (stderr, "Center pt is at %f %f %f\n", cx, cy, cz);
  Vector3 v2 = Vector3 (), v3 = Vector3 (), v2_prev = Vector3 ();
  Vector3 v_init;

  // For cylinder
  //   v4 and v5: pts of bottom triangle of each quad forming height of cyl.
  //     These are also the 2 base verts of each bott base sector triangle.
  //   v6: center of bottom base (just solHeight away from top base's center)
  Vector3 v4 = Vector3 (), v5 = Vector3 ();
  Vector3 v6 = Vector3 (cx, cy, cz+solHeight);

  // t: radians, in parameterized form of circle eqn.
  // z: always cz, we're drawing on xy plane, z doesn't change.
  float x = 0, y = 0, z = cz, t = 0;
  
  const int TTL_DEGS = 360;
  //const int TTL_DEGS = 90;
  int last_term = TTL_DEGS;

  // Texture coordinates
  // 2D. z always 0. origin is lowerleft corner of texture img, by convention
  // unit_w = total width of texture image / 360 degrees * x_step
  float unit_w = (float) _tex_w / TTL_DEGS * x_step;
  Vector3 t2_raw = Vector3 (unit_w, _tex_h, 0),
    t3_raw = Vector3 (0, _tex_h, 0),
    t4_raw = Vector3 (0,0,0),
    t5_raw = Vector3 (unit_w, 0, 0);
  Vector3 t2 = Vector3 (t2_raw[0]/_tex_w, t2_raw[1]/_tex_h, 0),
    t3 = Vector3 (0, t3_raw[1]/_tex_h, 0),
    t4 = Vector3 (0, 0, 0),
    t5 = Vector3 (t5_raw[0]/_tex_w, 0, 0);

  // Draw a triangle each loop. Triangles form circle.
  // Each triangle's LHS base vert (v3) == prev triangle's RHS base vert (v2)
  for (int i = 0; i < last_term; i+=x_step)
  {
    // i is current degree, convert into radian.
    t = BasicMath::radian (i);

    x = cx + r * cos (t);
    y = cy + r * sin (t);

    // Nothing to draw yet
    // y-coord == 0 `.` draw circ on the XZ plane
    if (i == 0)
    {
      v3.set (0, x);
      v3.set (1, y);
      v3.set (2, z);

      // Save for last triangle
      v_init.set (0, x);
      v_init.set (1, y);
      v_init.set (2, z);
    }

    // Draw normally for all other triangles. Set LHS base vert v3 to be same
    //   as prev triangle's RHS base vert v2
    else
    {
      // 1st iteration will draw the triangle with v3 computed in 0th iteratn,
      //   so don't reset v3. All later triangles' v3 should be the v2 of the
      //   previous triangle.
      // Assumption: v2 values have not changed since the prev triangle in
      //   last iter.
      if (i != x_step)
      {
        v3.set (0, v2_prev [0]);
        v3.set (1, v2_prev [1]);
        v3.set (2, v2_prev [2]);
      }

      if  (i != last_term - x_step)
      {
        v2.set (0, x);
        v2.set (1, y);
        v2.set (2, z);

        v2_prev.set (0, x);
        v2_prev.set (1, y);
        v2_prev.set (2, z);
      }
      // For last iteration, draw the last triangle closing in, RHS vert
      //   overlapping w/ the LHS vert of the 0th triangle.
      else
      {
        v2.set (0, v_init[0]);
        v2.set (1, v_init[1]);
        v2.set (2, v_init[2]);
      }

      v1.set (0, cx);
      v1.set (1, cy);
      v1.set (2, z);

//fprintf (stderr, "Drawing (%f %f %f) (%f %f %f) (%f %f %f)\n",
//  v1[0], v1[1], v1[2], v2[0], v2[1], v2[2], v3[0], v3[1], v3[2]);
//fprintf (stderr, "Heights are same? %f %f %f %f %f\n",
//  v1[2], v2[2], v3[2], cz, z);
//fprintf (stderr, "v2 %f %f %f v3 %f %f %f\n", v2[0], v2[1], v2[2], v3[0],
//  v3[1], v3[2]);

      drawTri (v1, v2, v3);
      //drawTri (v1, v3, v2);

      // Cylinder
      // draw the height face of a cyl, using 2 base pts of this triangle as
      //   the quad's top and bottom x y coords. Use solHeight as the diff
      //   btwn top and bottom bases.
      // We'll draw triangles instead of quads to make it consistent.
      if (isCyl == 1)
      {
        // Height of cyl

        // v4 is lowerleft corner of quad, right below v3. Add z for height
        v4.set (0, v3[0]);
        v4.set (1, v3[1]);
        v4.set (2, cz + solHeight);

        // v5 is lowerright corner of quad, right below v2. Add z for height
        v5.set (0, v2[0]);
        v5.set (1, v2[1]);
        v5.set (2, cz + solHeight);


        // If want texture, do glTexCoord2f() before glVertex3f() to define
        //   a texture coord (s,t,r,q) for the next vertex3f.
        if (doTexture)
        {
          // For i == 0, just use the initial vals. For subsequent ones,
          //   Increment each coord in horizontal direction. Vertical and z
          //   stays same, just shift to the right.
          if (i != 0)
          {
            // upperleft corner of upper triangle
            t3_raw.set (0, t3_raw[0] + unit_w);
            t3.set (0, t3_raw[0]/_tex_w);
            // lowerleft corner of upper triangle
            t4_raw.set (0, t4_raw[0] + unit_w);
            t4.set (0, t4_raw[0]/_tex_w);
            // upperright corner of upper triangle
            t2_raw.set (0, t2_raw[0] + unit_w);
            t2.set (0, t2_raw[0]/_tex_w);

            // Debug for texture coord increment to get the right triangles
            //fprintf (stderr, "%f %f %f, %f %f %f, %f %f %f\n",
            //  t3[0], t3[1], t3[2], t2[0], t2[1], t2[2], t4[0], t4[1], t4[2]);
          }

          drawTri_tex (v3,v2,v4, t3, t2, t4);
        }
        else
        {
          // Top triangle of quad
          drawTri (v3, v2, v4);
          //drawTri (v3, v4, v2);
        }

        // Debug. makes sure that drawTri didn't change its params' values
        //fprintf (stderr, "v2 %f %f %f v3 %f %f %f\n", v5[0], v5[1], v5[2],
        //  v4[0], v4[1], v4[2]);


        if (doTexture)
        {
          // t2 and t4 have been updated for upper triangle already, only 
          //   need to update t5.
          if (i != 0)
          {
            t5_raw.set (0, t5_raw[0] + unit_w);
            t5.set (0, t5_raw[0]/_tex_w);

            // Debug for texture coord increment to get the right triangles
            //fprintf (stderr, "%f %f %f, %f %f %f, %f %f %f\n",
            //  t2[0], t2[1], t2[2], t5[0], t5[1], t5[2], t4[0], t4[1], t4[2]);
          }

          drawTri_tex (v2,v5,v4, t2, t5, t4);
        }
        else
        {
          // Bott triangle of quad
          drawTri (v2, v5, v4);
          //drawTri (v2, v4, v5);
        }


        // Bottom base
        // v6 is center of bottom base
        v6.set (0, v1[0]);
        v6.set (1, v1[1]);
        v6.set (2, v1[2] + solHeight);

        drawTri (v6, v4, v5);
        //drawTri (v6, v5, v4);
      }

      // Cone
      else if (isCyl == 2)
      {
        // Same pos as center, just diff height. Add z for height
        v4.set (0, v1[0]);
        v4.set (1, v1[1]);
        v4.set (0, cz - solHeight);

        //drawTri (v2, v4, v3);
        drawTri (v2, v3, v4);
      }
    }

    //fprintf (stderr, "i %f v1 (%f %f %f) v2 (%f %f %f) v3 (%f %f %f)\n",
    //  i, v1[0], v1[1], v1[2], v2[0], v2[1], v2[2], v3[0], v3[1], v3[2]);
  }
}



// cx: x-coord of center of circ to draw
// cy: z-coord of center of circ to draw
// r: radius of circle, or a of ellipse, if mode is specified for ellp
// b: b axis length of ellipse, if mode is specified for ellp
// x_step: How fine grain the calculation needs to be. 0 to 360
// mode: constant, circ or ellp
// Does NOT implement ellps drawing yet. Use circPath (ELLP) to get ellp path.
void Shape::drawCircVert (double cx, double cy, double cz, double r, double b,
  int x_step, int mode)
{
  // To draw triangles that form each sector of the circle
  //   v1 is always the center of the circle. v2 and v3 go CCW, so the RHS
  //   and LHS verts on the isoceles triangle's base, respectively
  Vector3 v1 = Vector3 (cx, cy, cz);
  Vector3 v2 = Vector3 (), v3 = Vector3 (), v2_prev = Vector3 ();
  Vector3 v_init;

  // t: radians, in parameterized form of circle eqn.
  // z: always cz, we're drawing on xy plane, z doesn't change.
  //float x = 0, y = 0, z = cz, t = 0;
  float x = 0, y = cy, z = 0, t = 0;
  
  const int TTL_DEGS = 360;
  int last_term = TTL_DEGS;

  // Draw a triangle each loop. Triangles form circle.
  // Each triangle's LHS base vert (v3) == prev triangle's RHS base vert (v2)
  for (int i = 0; i < last_term; i+=x_step)
  {
    // i is current degree, convert into radian.
    t = BasicMath::radian (i);

    x = cx + r * cos (t);
    //y = cy + r * sin (t);
    z = cz + r * sin (t);

    // Nothing to draw yet
    // y-coord == 0 `.` draw circ on the XZ plane
    if (i == 0)
    {
      v3.set (0, x);
      v3.set (1, y);
      v3.set (2, z);

      // Save for last triangle
      v_init.set (0, x);
      v_init.set (1, y);
      v_init.set (2, z);
    }

    // Draw normally for all other triangles. Set LHS base vert v3 to be same
    //   as prev triangle's RHS base vert v2
    else
    {
      // 1st iteration will draw the triangle with v3 computed in 0th iteratn,
      //   so don't reset v3. All later triangles' v3 should be the v2 of the
      //   previous triangle.
      // Assumption: v2 values have not changed since the prev triangle in
      //   last iter.
      if (i != x_step)
      {
        v3.set (0, v2_prev [0]);
        v3.set (1, v2_prev [1]);
        v3.set (2, v2_prev [2]);
      }

      if  (i != last_term - x_step)
      {
        v2.set (0, x);
        v2.set (1, y);
        v2.set (2, z);

        v2_prev.set (0, x);
        v2_prev.set (1, y);
        v2_prev.set (2, z);
      }
      // For last iteration, draw the last triangle closing in, RHS vert
      //   overlapping w/ the LHS vert of the 0th triangle.
      else
      {
        v2.set (0, v_init[0]);
        v2.set (1, v_init[1]);
        v2.set (2, v_init[2]);
      }

      drawTri (v1, v2, v3);
    }

    //fprintf (stderr, "i %f v1 (%f %f %f) v2 (%f %f %f) v3 (%f %f %f)\n",
    //  i, v1[0], v1[1], v1[2], v2[0], v2[1], v2[2], v3[0], v3[1], v3[2]);
  }
}


// Parameters:
//   v1, v2, v3: 3 pts making up the triangle to draw.
//   t1, t2, t3: 3 pts, in texture coordinates (s,t,0)
void Shape::drawTri_tex (Vector3& v1, Vector3& v2, Vector3& v3,
  Vector3& t1, Vector3& t2, Vector3& t3)
{
  // textures usually want white on geometry
  Vector3 white = Vector3 (1, 1, 1);

  glTexCoord2f (t1[0], t1[1]);
  drawVert_clr_norm (v1, white, v1);
  glTexCoord2f (t2[0], t2[1]);
  drawVert_clr_norm (v2, white, v2);
  glTexCoord2f (t3[0], t3[1]);
  drawVert_clr_norm (v3, white, v3);
  num_faces++;
}

// Draw a triangle formed by the 3 vertices given.
void Shape::drawTri (Vector3& v1, Vector3& v2, Vector3& v3)
{
  drawVert (v1);
  drawVert (v2);
  drawVert (v3);
  num_faces++;
}

void Shape::drawTri_clr (Vector3& v1, Vector3& v2, Vector3& v3,
  Vector3& color1, Vector3& color2, Vector3& color3)
{
  drawVert_clr_norm (v1, color1, v1);
  drawVert_clr_norm (v2, color2, v2);
  drawVert_clr_norm (v3, color3, v3);
  num_faces++;
}

void Shape::drawTri_clr_norm (Vector3& v1, Vector3& v2, Vector3& v3,
  Vector3& color1, Vector3& color2, Vector3& color3,
  Vector3& norm1, Vector3& norm2, Vector3& norm3)
{
  drawVert_clr_norm (v1, color1, norm1);
  drawVert_clr_norm (v2, color2, norm2);
  drawVert_clr_norm (v3, color3, norm3);
  num_faces++;
}

void Shape::drawQuad (Vector3& v1, Vector3& v2, Vector3& v3, Vector3& v4)
{
  drawVert (v1);
  drawVert (v2);
  drawVert (v3);
  drawVert (v4);
  num_faces++;
}

// Draw the vertex passed in
void Shape::drawVert (Vector3& vert)
{
  float r = 0, g = 0, b = 0;

  // set color to magenta
  //glColor3f(1.0 / NUM_TRIS * i, 0.0, 1.0 / NUM_TRIS * i);
  //glColor3f(1.0, 0.0, 1.0);
  
  r = (num_faces * 1 % 256) / 255.0;
  g = (num_faces * 25 % 256) / 255.0;
  b = (num_faces * 50 % 256) / 255.0;

  Vector3 color = Vector3 (r, g, b);

  // When no normal is specified, just pass in the pt itself to be normal
  drawVert_clr_norm (vert, color, vert);
}

// Lets user pass in a color value
void Shape::drawVert_clr_norm (Vector3& vert, Vector3& color, Vector3& norm)
{

  float x = 0, y = 0, z = 0;
  
  glColor3f(color[0], color[1], color[2]);

  // These are normals of the actual icosahedron, not the sphere.
  // Set normals, 1 normal for each triangle face
  //for (int j = 0; j < 3; j++)
  //{
    // Because the normal of all 3 verts of a triangle should be the same,
    //   only need to calc once, to set the normal for all 3 vecs
    // Normal of a surface is the normalized cross product btwn 2 edges
    //   on that surface
    // Subtract each jth component (x y z) to get the vector
  //  diff1.set (j, vertData_arr [triIndices [i][0]] [j] -
  //    vertData_arr [triIndices [i][1]] [j]);
  //  diff2.set (j, vertData_arr [triIndices [i][1]] [j] -
  //    vertData_arr [triIndices [i][2]] [j]);
  //}
  // Calc cross prod of a vec connecting 2 edges
  //diff1.cross (diff2, crossed);
  // Normalize cross prod
  //crossed.norm (crossed);
  //glNormal3f (crossed[0], crossed[1], crossed[2]);
  //


  

  // Rotate - doesn't work, distorted shapes. Maybe because of Matrix4::rotate?
  //   Image is distorted even when _rot* are not set. And what should they be
  //   defaulted to anyway, identity? 0s? a combo of both?
  /*
  Matrix4 rotMat = Matrix4 ();
  rotMat.rotate (_rotx, _roty, _rotz, _rotw);
  // DO NOT DO THIS!! arg should stay constant. This will cause problems in
  //   drawCirc(), when v1 v2 v3 v4 v5 are reused!!!
  rotMat.mul (vert, vert);
  */

  // Scale
  //Matrix4 scaleMat = Matrix4 ();
  //scaleMat.scale (_scalex, _scaley, _scalez);
  // DO NOT DO THIS!! arg should stay constant. This will cause problems in
  //   drawCirc(), when v1 v2 v3 v4 v5 are reused!!!
  //scaleMat.mul (vert, vert);

  // Testing multiplication
  //Vector4 testVec4 = Vector4 (vert[0], vert[1], vert[2], 0);
  //scaleMat.mul (testVec4, testVec4);
  //testVec4.dehom (testVec4);
  //Vector3 testVec3 = Vector3 (testVec4[0], testVec4[1], testVec4[2]);

  // Translate
  /*
  Matrix4 transMat = Matrix4 ();
  transMat.translate (20, 0, 0);
  // DO NOT DO THIS!! arg should stay constant. This will cause problems in
  //   drawCirc(), when v1 v2 v3 v4 v5 are reused!!!
  transMat.mul (vert, vert);
  */


  // Normals of the approximated sphere
  // Each normal pts in the same dir as the vec from origin to the
  //   corresponding vertex.
  // Testing multiplication
  //x = testVec3[0];
  //y = testVec3[1];
  //z = testVec3[2];
  x = vert[0];
  y = vert[1];
  z = vert[2];

  // Do not change the norm vec itself. keep it CONST.
  Vector3 norm_tmp = Vector3 (norm[0], norm[1], norm[2]);
  norm.norm (norm_tmp);
  glNormal3f (norm_tmp[0], norm_tmp[1], norm_tmp[2]);
  // Draw verts
  // i = ith triangle, 0 1 2 = vertex number in ith triangle (defined in
  //   triIndices). Last [0] is the index in vertData, we're just taking the
  //   ptr &arr[0] so can pass the entire array to glVertex3fv, which takes
  //   an array of 3 floats
  glVertex3f (x, y, z);

}







