/*
Copyright (C) 2006 So Yamaoka

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#include <cstdio>
#include <cstdlib>

#include "shader.h"


// ------------------------------------------------------------- ctor, dtor --

// ctor
Shader::Shader ()
{
}

// Parameters:
//   vert: file name of vertex shader program
//   frag: file name of fragment shader program
Shader::Shader(const char *vert, const char *frag, bool isFile)
{
  if (isFile)
  {
    // Read file and save file content into a char array (ptr to it returned)
    char* vv = read(vert);
    char* vf = read(frag);

    setup(vv, vf);
    delete[] vv;
    delete[] vf;
  }
  else
    setup(vert, frag);
}

Shader::~Shader()
{
  glDeleteObjectARB(pid);
  //fprintf (stderr, "===================================================\n");
}


// ---------------------------------------------------------------------------

// Called by ctor to read in shader program files
// Return: pointer to char array containing the shader file loaded.
char* Shader::read(const char *filename)
{
  char* shaderFile = 0;
  
  // Open shader program
  FILE* fp = fopen(filename, "rb");
  if(!fp)
  {
    fprintf(stderr, "File doesn't exist (%s)\n", filename);
    std::exit(-1);
  }

  // Get file size
  fseek(fp, 0, SEEK_END);
  long size = ftell(fp);
  rewind(fp);

  // Alloc memory - will be deleted while setting the shader up
  // Put on heap so that when pointer returns, this won't be out of scope (
  //   would be if it's on stack. .`. put on heap.)
  shaderFile = new char[size+1];

  // Copy file to shaderFile
  fread (shaderFile, sizeof(char), size, fp);
  shaderFile [size] = '\0'; // eliminate the garbage at EOF
  fclose (fp);

  fprintf(stdout, "shaderfile = %p\n", shaderFile);
  return shaderFile;
}

// Called by ctor
void Shader::setup(const char *vs, const char *fs)
{
  //fprintf(stdout, "vs, fs = %p, %p\n", vs, fs);
  
  //GLuint vid = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
  //GLuint fid = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
  GLhandleARB vid = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
  GLhandleARB fid = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
  //fprintf(stderr, "vid %d fid %d\n", vid, fid);
  
  glShaderSourceARB(vid, 1, &vs, 0);
  glShaderSourceARB(fid, 1, &fs, 0);
  
  glCompileShaderARB(vid);
  glCompileShaderARB(fid);
  char glslLog[1024];
  GLsizei glslLogSize;
  glGetInfoLogARB(vid, 1024, &glslLogSize, glslLog);
  if (glslLogSize)
    printf("vertex program log: %s\n", glslLog);
  glGetInfoLogARB(fid, 1024, &glslLogSize, glslLog);
  if (glslLogSize)
    printf("fragment program log: %s\n", glslLog);

  pid = glCreateProgramObjectARB();
  //fprintf(stderr, "this is the problem: pid %d\n", pid);
  glAttachObjectARB(pid, vid);
  //fprintf (stderr, "glAttach: ");
  //printGLerr ();
  glAttachObjectARB(pid, fid);
  //fprintf (stderr, "glAttach: ");
  //printGLerr ();


  // delete shader objects as soon as they have been attached to a program
  glDeleteObjectARB(vid);
  glDeleteObjectARB(fid);
  
  glLinkProgramARB(pid);
  printGLerr ();

}


// ---------------------------------------- Fns that can be called from app --

void Shader::bind()
{
  //fprintf (stderr, "Binding shader %d\n", pid);
  glUseProgramObjectARB(pid);
}

void Shader::unbind()
{
  //fprintf (stderr, "Unbinding shader\n");
  glUseProgramObjectARB(0);
}

void Shader::printLog(const char* tag)
{
  char glslLog[1024];
  GLsizei glslLogSize;
  glGetInfoLogARB(pid, 1024, &glslLogSize, glslLog);
  if(glslLogSize>0)
    fprintf(stderr, "%s:shader error log: %s\n", tag, glslLog);
  else
    fprintf(stdout, "%s:shader all ok.\n", tag);

  glGetProgramInfoLog (pid, 1024, &glslLogSize, glslLog);
  if(glslLogSize>0)
    fprintf(stderr, "%s:shader error log: %s\n", tag, glslLog);
  else
    fprintf(stdout, "%s:shader all ok.\n", tag);

}


void Shader::printGLerr ()
{
  //fprintf (stderr, "Error nubmer: %d (want %d) **************** \n",
  //  glGetError (), GL_NO_ERROR);
  GLenum errnum = glGetError ();
  switch (errnum)
  {
    case GL_NO_ERROR:
      fprintf (stderr, "GL_NO_ERROR\n");
      break;
    case GL_INVALID_ENUM:
      fprintf (stderr, "GL_INVALID_ENUM\n");
      break;
    case GL_INVALID_VALUE:
      fprintf (stderr, "GL_INVALID_VALUE\n");
      break;
    case GL_INVALID_OPERATION:
      fprintf (stderr, "GL_INVALID_OPERATION\n");
      break;
    case GL_STACK_OVERFLOW:
      fprintf (stderr, "GL_STACK_OVERFLOW\n");
      break;
    case GL_STACK_UNDERFLOW:
      fprintf (stderr, "GL_STACK_UNDERFLOW\n");
      break;
    case GL_OUT_OF_MEMORY:
      fprintf (stderr, "GL_OUT_OF_MEMORY\n");
      break;
    case GL_TABLE_TOO_LARGE:
      fprintf (stderr, "GL_TABLE_TOO_LARGE\n");
      break;
    default:
      fprintf (stderr, "Some other error...?\n");
      break;
  }
}

